#!/bin/sh
# \
exec wish "$0" "$@"

proc query_init {} {
	global Query
	load libpgtcl.so
	set Query(db) ""
	set Query(conn) ""
}

proc query_status_chk { handle } {
	set error [pg_result $handle -error]
	if {$error != ""} {
		set result [pg_result $handle -status]
		bgerror "QUERY: $result --> $error"
	}
}

proc query_opendb { db } {
	global Query
	if {$Query(db) == $db} { return }
	if {$Query(conn) != ""} {
		pg_disconnect $Query(conn)
	}
	set Query(conn) [pg_connect $db]
}

proc query_closedb { db } {
	global Query
	pg_disconnect $Query(conn)
	set Query(conn) ""
}

proc query { q args } {
	global Query
	proc $q { args } "return \[QueryProc $q \$args]" 
	set Query($q,subq) {}
	set Query($q,base) {}
	set Query($q,-select) ""
	set Query($q,-insert) ""
	set Query($q,-update) ""
	set Query($q,-delete) ""
	set Query($q,-class) $q 
	set Query($q,-multiple) 0
	set Query($q,handle) ""
	set attrs [Query_attributes $q]
	set first [lindex [lindex $attrs 0] 0]
	set Query($q,-order) $first
	set Query($q,-idattr) ""
	#set Query($q,-idattr) $first
	if { [llength $args] } {
		eval $q config $args
	}
	return $q
}

proc QueryProc { q qArgs } {
	global Query
	set cmd [lindex $qArgs 0]
	set cmdArgs [lrange $qArgs 1 end]
	switch -glob -- $cmd {
	config* {
		if {[llength $cmdArgs] == 1} {
			return $Query($q,[lindex $cmdArgs 0])
		} else {
			array set opts $cmdArgs
			foreach opt [array names opts] {
				switch -glob -- $opt {
				-select -
				-update -
				-insert -
				-delete -
				-multiple - 
				-class -
				-order -
				-idattr {
					set Query($q,$opt) $opts($opt)
				}
				default {
					error "option unknown, valid options: -select, \
						-update, -delete,-multiple"
				}
				}
			}
		}
	}
	select {
		eval Query_select $q $cmdArgs
	}
	update -
	insert -
	delete {
		eval Query_$cmd $q
	}
	set {
		if {[llength $cmdArgs] == 0} {
			return $Query($q,ntup)
		}
		switch -glob -- [lindex $cmdArgs 0] {
		+1 { 
			if {$Query($q,ntup) < [expr $Query($q,maxtup)-1]} {
				incr Query($q,ntup)
			}
		}
		-1 {
			if {$Query($q,ntup) > 0} {
				incr Query($q,ntup) -1
			}
		}
		end {
			set Query($q,ntup) [expr $Query($q,maxtup)-1]
		}
		new {
			set Query($q,ntup) $Query($q,maxtup)
		}
		default {
			if {[lindex $cmdArgs 0] < $Query($q,maxtup)} {
				set Query($q,ntup) [lindex $cmdArgs 0]
			}
		}
		}
		global $q
		if {!$Query($q,-multiple)} {
			pg_result $Query($q,handle) -tupleArray $Query($q,ntup) $q 
		}
	}
	maxtuple {
		return $Query($q,maxtup)
	}
	add {
		if {[llength $cmdArgs] == 1} {
			set idx $Query([lindex $cmdArgs 0],-idattr)
		} else {
			set idx [lindex $cmdArgs 1]
		}
		lappend Query($q,subq) [list $Query([lindex $cmdArgs 0],-class) \
			$idx]
	}
	clear {
		global $q
		upvar #0 $q v
		foreach n [array names $q] {
			set v($n) {}
		}
		pg_result $Query($q,handle) -clear
		set Query($q,handle) ""		
	}
	attributes {
		return [Query_attributes $q]
	}
	default {
		bgerror "cmd desconhecido: $cmd"
	}
	}
}

proc Query_select { q  args} {
	global Query
	if {$Query($q,-select) != ""} {
		set qs [subst $Query($q,-select)]
	} else {
		if {![info exists Query($q,-class)]} {
			error "class not specified"
		}
		set b $Query($q,-class)
		set qs "select b.oid,* from $b b"
		set i 0
		foreach sq $Query($q,subq) {
			append qs ",[lindex $sq 0] c$i"
			incr i
		}
		set i 0
		foreach sq $Query($q,subq) {
			if {$i > 0} {
				append qs " and "
			} else {
				append qs " where "
			}
			append qs "b.[lindex $sq 1] = c$i.[lindex $sq 1]"
			incr i	
		}
# append where clauses in "where"
		if {[llength $args]} {	
			if {$i == 0} {
				append qs " where "
			} else {
				append qs " and "
			}
			upvar #0 $q v
			foreach {op attr} $args {
				switch -- $op {
				-like {
						append qs "$attr like '$v($attr)%'"
					}
				-eq {
						append qs "$attr = '$v($attr)'"
					}
				-sql {
						append qs $attr
					}
				}
			}
		}
		if {$Query($q,-order) != ""} {
			append qs " order by $Query($q,-order)"
		}
	}
#puts "SELECT: $qs"
# if another query pending, clear it before redoing 
	if {$Query($q,handle) != ""} {
#puts "pg_result $Query($q,handle) -clear"
		pg_result $Query($q,handle) -clear
	}
	set Query($q,handle) [pg_exec $Query(conn) $qs]
#puts "STATUS: [pg_result $Query($q,handle) -status]"
	query_status_chk $Query($q,handle)
	set Query($q,ntup) 0
	global $q
	set Query($q,maxtup) [pg_result $Query($q,handle) -numTuples]
	catch {unset $q}
	if {$Query($q,-multiple)} {
		pg_result $Query($q,handle) -assign $q
	} else {
		pg_result $Query($q,handle) -tupleArray 0 $q 
	}
}

proc Query_attributes { q } {
	global Query
	if {![info exists Query($q,attrs)]} {
   		set qry "select a.attname,t.typname,a.attlen,a.atttypmod,
			a.attnotnull from \
       		pg_attribute a, pg_class c, pg_type t \
       		where a.attrelid = c.oid \
       		and c.relname = '$Query($q,-class)' \
       		and a.attnum > 0 \
       		and a.atttypid = t.oid order by a.attnum"
   		set Query($q,attrs) {}
		set Query($q,attrnames) {}
		pg_select $Query(conn) $qry v { 
			if {$v(attlen) == -1} {
				set len [expr $v(atttypmod) - 4]
			} else {
				set len $v(attlen)
			}
			lappend Query($q,attrnames) $v(attname) 
			lappend Query($q,attrs) \
				[list $v(attname) $len $v(typname) \
				[string compare $v(attnotnull) "f"]] 
		}
	}
	return $Query($q,attrs)
}

proc Query_insert { q } {
	global Query
	if {$Query($q,-insert) != ""} {
		set qqs [subst $Query($q,-select)]
	} else {
		if {![info exists Query($q,-class)]} {
			error "class not specified"
		}
		upvar #0 $q v
		set attributes [Query_attributes $q]
		set b $Query($q,-class)
		set qqs "insert into $b\(" 
		set qs ") values ("
		set i 0
		foreach atlist $attributes {
			set at [lindex $atlist 0]
			if {$i > 0} {
				append qs ,
				append qqs ,
			}
			if {$Query($q,-multiple)} {
				set k $Query($q,ntup)
				if {[info exists v($k,$at)]} {
					append qs "'$v($k,$at)'"
				} else {
					append qs ''
				}
			} else {
				if {[info exists v($at)]} {
					append qs "'$v($at)'"
				} else {
					append qs ''
				}
			}
			append qqs $at
			incr i
		}
		append qs )
		append qqs $qs
	}
#puts "INSERT: $qqs"
	set handle [pg_exec $Query(conn) $qqs]
	query_status_chk $handle
	pg_result $handle -clear
}

proc Query_update { q } {
	global Query
	if {$Query($q,-update) != ""} {
		set qs [subst $Query($q,-select)]
	} else {
		if {![info exists Query($q,-class)]} {
			error "class not specified"
		}
		if {$Query($q,-idattr) != ""} {
			set qual $Query($q,-idattr)
		} else {
			set qual oid
		}
		upvar #0 $q v
		set b $Query($q,-class)
		set qs "update $b set "
		set i 0
		Query_attributes $q
		foreach at $Query($q,attrnames) {
			if {$i > 0} {
				append qs ,
			}
			if {$Query($q,-multiple)} {
				set k $Query($q,ntup)
				append qs "$at = '$v($k,$at)'"
			} else {
				append qs "$at = '$v($at)'"
			}
			incr i
		}
		if {$Query($q,-multiple)} {
			set k $Query($q,ntup)
			append qs " where $qual = '$v($k,$qual)'"
		} else {
			append qs " where $qual = '$v($qual)'"
		}
	}
puts "UPDATE: $qs"
	set handle [pg_exec $Query(conn) $qs]
#puts "status: [pg_result $handle -status]"
	query_status_chk $handle
	pg_result $handle -clear
}

proc Query_delete { q } {
	global Query
	if {$Query($q,-update) != ""} {
		set qs [subst $Query($q,-select)]
	} else {
		if {![info exists Query($q,-class)]} {
			error "class not specified"
		}
		if {$Query($q,-idattr) != ""} {
			set qual $Query($q,-idattr)
		} else {
			set qual oid
		}
#puts "DELETE qual: $qual"
		upvar #0 $q v
		set b $Query($q,-class)
		set qs "delete from $b "
		if {$Query($q,-multiple)} {
			set k $Query($q,ntup)
			append qs "where $qual = $v($k,$qual)"
		} else {
			append qs "where $qual = $v($qual)"
		}
	}
puts "DELETE: $qs"
	set handle [pg_exec $Query(conn) $qs]
#puts "status: [pg_result $handle -status]"
	query_status_chk $handle
	pg_result $handle -clear
}

